/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2014 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include <QtCore/QCoreApplication>
#include <QTextStream>
#include <QDir>
#include <QFile>
#include <QSettings>
#include <QCryptographicHash>
#include <QtSql/QSqlQuery>
#include <QtSql/QSqlQueryModel>
#include <QtSql/QSqlError>
#include <QtSql/QSqlRecord>
#include <QCommandLineParser>
#include <QSslSocket>

// The Qt Cryptographic Architecture (QCA) from http://delta.affinix.com/qca/ has to be installed to be able to compile
#ifndef Q_OS_WIN32
    #include <QtCrypto>
#endif

#include "main.h"
#include "../main.h"
#include "configure.h"
#include "../kembossd/sql.h"
#include "../kembossd/ksshared.h"

int main(int argc, char *argv[])
{
    QCoreApplication app(argc, argv);

    QCoreApplication::setOrganizationName("kemboss.org");
    QCoreApplication::setOrganizationDomain("kemboss.org");
    QCoreApplication::setApplicationName("kembossd-admin");
    QCoreApplication::setApplicationVersion(QString(KEMBOSS_SERVER_MAJOR_VERSION)+"."+QString(KEMBOSS_SERVER_MINOR_VERSION)+"."+QString(KEMBOSS_SERVER_MINIMUM_VERSION));

    QCommandLineParser parser;
    parser.setApplicationDescription("Administration program for the EMBOSS server kembossd.");
    parser.addHelpOption();
    parser.addVersionOption();

    // Process the actual command line arguments given by the user
    parser.process(app);

#ifndef Q_OS_WIN32
    QCA::Initializer init;
#endif

    QTextStream qstdout(stdout);

    configFile = kembossServer::getServerConfigFile();

    // Check for server configuration file
    qstdout << "Reading server configuration in " << configFile << endl;
    if (!QFile::exists(configFile)) {
        qstdout << "No server configuration found. Creating new file." << endl;
    }

    // Check database configuration
    ksql::checkDb();

    // Check if system supports SSL
    if (!QSslSocket::supportsSsl()) {
        qstdout << "Your system doesn't support SSL. Qt might have been compiled with SSL disabled or the OpenSSL libraries are not installed on Your system." << endl;
        exit(1);
    }

    receiveCommand();

    return app.exec();
}

void receiveCommand()
{
    QTextStream qstdout(stdout);
    QTextStream qstdin(stdin);

    qstdout << endl;
    qstdout << "What do You want to do:" << endl << endl;
    qstdout << "1 Configure server" << endl;
    qstdout << "2 Configure SSL connection" << endl;
    qstdout << "3 Configure SQL database" << endl;
    qstdout << "4 Configure installed EMBASSY packages" << endl;
    qstdout << "5 Enter new user" << endl;
    qstdout << "6 Delete server log entries older than x days" << endl;
    qstdout << "7 Delete cached input files that where not accessed for x days" << endl;
    qstdout << "8 Check input file cache" << endl;
    qstdout << "9 Print user list" << endl;
    qstdout << "v Version information" << endl;
    // qstdout << "11 Create statistics" << endl;

    qstdout << "x Quit server administration" << endl << endl;
    qstdout << "Please enter number: " << flush;

    QString command;
    command = qstdin.read(1);
    qstdout << endl;
    if (command == "x") {
        exit(0);
    } else if (command == "v") {
        displayVersion();
    } else if (command == "1") {
        configureServer(configFile);
    } else if (command == "2") {
        configureSSL(configFile);
    } else if (command == "3") {
        configureDb(configFile);
    } else if (command == "4") {
        configureEmbassyPackages(configFile);
    } else if (command == "5") {
        enterNewUser();
    } else if (command == "6") {
        deleteOldLogEntries();
    } else if (command == "7") {
        deleteOldInputFiles();
    } else if (command == "8") {
        checkInputFileCache();
    } else if (command == "9") {
        printUserList();
  /*  } else if (command == "s") {
        showStatistics();*/
    }

    receiveCommand();
}

int enterNewUser()
{
    QTextStream qstdout(stdout);
    QTextStream qstdin(stdin);

    qstdout << "Enter new user..." << endl;
    qstdout << "Enter user name: " << flush;
    QString name;
    name = qstdin.readLine();
    qstdout << "Enter group (optional): " << flush;
    QString group;
    group = qstdin.readLine();

    // The kemboss client hashes the password entered by the user using Sha1 and stores the resulting hash on the hard disk (so the plain text password is not saved on disk)
    // The password hash is send to the server. The server again hashes the received hash using Sha256 (so that if someone gains access to the password database, he is not able to do anything with it)
    // Here the entered password has first to be hashed using Sha1 and the resulting hash has to be hashed again using Sha256

#ifndef Q_OS_WIN32
    QCA::SecureArray pass;
    QCA::ConsolePrompt prompt;
    do {
        prompt.getHidden("Enter user password: ");
        prompt.waitForFinished();
        pass = prompt.result();
    } while (pass.isEmpty());
    QByteArray pw = kembossServer::SHA256hash(QCryptographicHash::hash(pass.toByteArray(), QCryptographicHash::Sha1));
#else
    QString pass = "";
    do {
        qstdout << "Enter user password: " << flush;

        enableEcho(false);
        pass = qstdin.readLine();
        enableEcho(true);

        qstdout << endl;

    } while (pass.isEmpty());
    QByteArray pw = kembossServer::SHA256hash(QCryptographicHash::hash(pass.toUtf8(), QCryptographicHash::Sha1));
#endif

    // A single ' in the password hash has to be escaped by '', otherwise the SQL statement checking the user is invalid
    pw.replace("'","''");

    qstdout << "Enter user email address (optional): " << flush;
    QString email;
    email = qstdin.readLine();

    ksql::newUser(name, group, pw, true, email);

    return(0);
}

bool printUserList()
{
    QTextStream qstdout(stdout);
    QTextStream qstdin(stdin);

    qstdout << "Save user list in file: " << flush;
    QString filename;
    filename = qstdin.readLine();

    QFile file(filename);
    if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
        qstdout << "Couldn't open file to save user list" << endl;
        return false;
    }
    QTextStream qfout(&file);
    qfout << "ID   Username    Group       Email               Created             Last_Login         \n";

    QSqlQueryModel result;
    bool next = true;

    {
        QSqlDatabase db;

        ksql::dbType type = ksql::checkDbDriver();

        db = QSqlDatabase::addDatabase(ksql::dbTypeStr(type), "kembossd");
        if (!ksql::connectToDb(&db, type))
            next = false;

        if (next)
            if (!db.isValid()) {
                kembossServer::msgStdErr("Error in adding the database to the list of database connections: " + db.lastError().text());
                next = false;
            }

        if (next) {
            QSqlQuery query(db);

            if (!query.exec("SELECT id, name, `group`, email, created, last_login FROM `user`")) {
                kembossServer::msgStdErr(query.lastError().text());
                kembossServer::msgStdErr(db.lastError().text());
                next = false;
            }

            if (next && !query.isActive()) {
                kembossServer::msgStdErr("The SQL statement has not been executed successfully.");
                next = false;
            }
            if (next && !query.isSelect()) {
                kembossServer::msgStdErr("The SQL statement is no SELECT statement, so no result can be retrieved.");
                next = false;
            }

            if (next) {
                result.setQuery(query);
            }
        }

        for (int j = 0; j < result.rowCount(); ++j) {
            qfout << result.data(result.index(j, 0)).toString().leftJustified(5); //id
            qfout << result.data(result.index(j, 1)).toString().leftJustified(12); //name
            qfout << result.data(result.index(j, 2)).toString().leftJustified(12); //group
            qfout << result.data(result.index(j, 3)).toString().leftJustified(20); //email
            qfout << result.data(result.index(j, 4)).toString().leftJustified(20); //created
            qfout << result.data(result.index(j, 5)).toString().leftJustified(19); //last_login
            qfout << "\n";
        }

        file.close();

        db.close();
    }

    QSqlDatabase::removeDatabase("kembossd");

    qstdout << "Saved user list to: " << filename << "\n" << endl;

    return true;
}

bool deleteOldLogEntries()
{
    QTextStream qstdout(stdout);
    QTextStream qstdin(stdin);

    qstdout << "Delete log entries older than (days): " << flush;
    QString days;
    days = qstdin.readLine();
    if (days.toInt() == 0) {
        qstdout << "Please enter an integer > 0" << flush;
        return false;
    }
    uint num;
    num = ksql::execSqlQuery("SELECT COUNT(id) FROM program_run WHERE DATEDIFF(CURRENT_DATE, run_at) > "+days,
                             "Error in deleting old log entries.", true).toInt();
    ksql::execSqlQuery("DELETE FROM program_run WHERE DATEDIFF(CURRENT_DATE, run_at) > "+days,
                       "Error in deleting old log entries.");
    qstdout << num << " log entries deleted" << endl;
    return true;
}

bool showStatistics()
{
 /*   QTextStream qstdout(stdout);
    qstdout << "Creating statistics..." << endl << endl;*/

    return true;
}

void displayVersion()
{
    QTextStream qstdout(stdout);
    qstdout << "kembossd version: " << QCoreApplication::applicationVersion() << endl;
    qstdout << "Qt version: " << QString(qVersion()) << endl;
    qstdout << "SSL version: " << QSslSocket::sslLibraryVersionString() << endl;
}

bool deleteOldInputFiles()
{
    QTextStream qstdout(stdout);
    QTextStream qstdin(stdin);

    qstdout << "Delete cached input files that where not accessed for (days): " << flush;
    QString days;
    days = qstdin.readLine();
    if (days.toInt() == 0) {
        qstdout << "Please enter an integer > 0" << flush;
        return false;
    }

    int removed = 0;

    QSettings settings(configFile, QSettings::IniFormat);
    QString inputFileDir = settings.value("UserDir", "").toString()+"/";

    {
        QSqlQueryModel result;
        QSqlDatabase db;

        ksql::dbType type = ksql::checkDbDriver();

        db = QSqlDatabase::addDatabase(ksql::dbTypeStr(type), "kembossd");
        if (!ksql::connectToDb(&db, type))
            return false;

        if (!db.isValid()) {
            kembossServer::msgStdErr("Error in adding the database to the list of database connections: " + db.lastError().text());
            return false;
        }

        QSqlQuery query(db);

        // Get all files that where not accessed for x days from the database
        if (!query.exec("SELECT id, user, file FROM input_files WHERE DATEDIFF(CURRENT_DATE, last_accessed) > "+days)) {
            kembossServer::msgStdErr(query.lastError().text());
            kembossServer::msgStdErr(db.lastError().text());
            return false;
        }
        if (!query.isActive()) {
            kembossServer::msgStdErr("The SQL statement has not been executed successfully.");
            return false;
        }
        if (!query.isSelect()) {
            kembossServer::msgStdErr("The SQL statement is no SELECT statement, so no result can be retrieved.");
            return false;
        }

        result.setQuery(query);

        for (int j = 0; j < result.rowCount(); ++j) {
            // Delete the file from disk
            if (QFile::remove(inputFileDir+result.data(result.index(j, 1)).toString()+"/input_files/"+result.data(result.index(j, 2)).toString())) {
                removed++;
                qstdout << "Deleted: " << result.data(result.index(j, 1)).toString() << "/input_files/" << result.data(result.index(j, 2)).toString() << endl;
            } else {
                qstdout << "Unable to delete file: " << result.data(result.index(j, 1)).toString() << "/input_files/" << result.data(result.index(j, 2)).toString() << endl;
            }
            // Delete the database entry
            ksql::execSqlQuery("DELETE FROM input_files WHERE id="+result.data(result.index(j, 0)).toString(),
                               "Error in deleting input file entry.");
        }

        db.close();
    }

    QSqlDatabase::removeDatabase("kembossd");

    qstdout << "Deleted " << removed << " files" << endl;

    return true;
}

bool checkInputFileCache()
{
    QTextStream qstdout(stdout);
    qstdout << "Checking input file cache..." << endl;

    QSettings settings(configFile, QSettings::IniFormat);
    QString inputFileDir = settings.value("UserDir", "").toString()+"/";

    int removed = 0;

    {
        QSqlQueryModel result;
        QSqlDatabase db;

        ksql::dbType type = ksql::checkDbDriver();

        db = QSqlDatabase::addDatabase(ksql::dbTypeStr(type), "kembossd");
        if (!ksql::connectToDb(&db, type))
            return false;

        if (!db.isValid()) {
            kembossServer::msgStdErr("Error in adding the database to the list of database connections: " + db.lastError().text());
            return false;
        }

        QSqlQuery query(db);
        QSqlQuery singleQuery(db);

        // Delete database entries where the file doesn't exist on disk any more
        // Get all files
        if (!query.exec("SELECT input_files.id, user.name, file FROM user, input_files WHERE input_files.user = user.id")) {
            kembossServer::msgStdErr(query.lastError().text());
            kembossServer::msgStdErr(db.lastError().text());
            return false;
        }

        if (!query.isActive()) {
            kembossServer::msgStdErr("The SQL statement has not been executed successfully.");
            return false;
        }
        if (!query.isSelect()) {
            kembossServer::msgStdErr("The SQL statement is no SELECT statement, so no result can be retrieved.");
            return false;
        }

        result.setQuery(query);

        for (int j = 0; j < result.rowCount(); ++j) {
            if (!QFile::exists(inputFileDir+result.record(j).value("name").toString()+"/input_files/"+result.record(j).value("file").toString())) {
                qstdout << "File " << inputFileDir+result.record(j).value("name").toString()+"/input_files/"+result.record(j).value("file").toString() << " doesn't exist in filecache. Deleting database entry." << endl;
                // Delete the database entry
                if (!singleQuery.exec("DELETE FROM input_files WHERE id="+result.record(j).value("id").toString()))
                   qstdout << "Error in deleting input file entry." << endl;
                else
                   removed++;
            }
        }

        db.close();
    }

    QSqlDatabase::removeDatabase("kembossd");

    qstdout << "Deleted " << removed << " entries in the database where the file doesn't exist on disk any more" << endl << endl;


    // Check for files on the disk that don't exist in the SQL database and delete them
    // Get all user dirs
    QDir dir(inputFileDir);
    QStringList userList;
    userList = dir.entryList(QDir::Dirs | QDir::NoDotAndDotDot);

    // Search all input files of all users and check if they exist in the database
    removed = 0;
    QStringList fileList;
    foreach (QString userName, userList) {
        dir.setPath(inputFileDir+userName+"/input_files/");
        fileList = dir.entryList(QDir::Files);
        foreach(QString filename, fileList) {
            if (ksql::execSqlQuery("SELECT input_files.id FROM input_files, user WHERE user.id = input_files.user AND user.name='"+userName+"' AND file='"+filename+"'",
                                   "Error in checking the file ID.", true).toInt() == 0) {
                // Database entry does not exist
                qstdout << "Deleting file: " << inputFileDir << userName << "/input_files/" << filename << " from filecache" << endl;
                if (QFile::remove(inputFileDir+userName+"/input_files/"+filename))
                    removed++;
            }
        }
    }

    qstdout << "Deleted " << removed << " input files that don't exist in the database" << endl;

    return true;
}
