/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2019 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#ifndef KEMBOSSSERVER_H
#define KEMBOSSSERVER_H

#include <QString>
#include <QTcpServer>
#include <QList>
#include <QStringList>
#include <QDir>
#include <QThread>
#include "kthread.h"
#include "kinputThread.h"
#include "../shared.h"

class KembossServer : public QTcpServer
{
    Q_OBJECT

public:
    KembossServer(QObject *parent = nullptr);
    ~KembossServer();
    bool loadEmbossProgramList(const QString &filename);

    void setServerStatus(const kemboss::serverStatus status) { cserverStatus = status; }
    void setAcceptAllUsers(const bool accept) { cacceptAllUsers = accept; }
    bool setEmbossFileDir(const QString &dir);
    void setEmbossBinDir(const QString &dir) { cembossBinDir = dir; if (!cembossBinDir.endsWith("/")) cembossBinDir += "/"; }
    void setWelcomeMessage(const QString &msg) { cwelcomeMessage = msg; }
    void setMaxEmbossProcessTime(const uint time) { cmaxEmbossProcessTime = time; }
    void setMaxThreadNumber(const uint num) { cmaxThreadNumber = num; }
    void setPhylipnewInstalled(const bool installed) { cphylipnewInstalled = installed; }
    void setLogfile(const QString &filename) { clogfile = filename; }
    void setPrivateKeyFile(const QString &filename);
    void setCACertsPath(const QString &path);
    void setLocalCertFile(const QString &filename);

    kemboss::serverStatus getServerStatus() { return cserverStatus; }
    bool getAcceptAllUsers() { return cacceptAllUsers; }
    QString getEmbossFileDir() { return cembossFileDir; }
    QString getEmbossBinDir() { return cembossBinDir; }
    QString getWelcomeMessage() { return cwelcomeMessage; }
    uint getMaxEmbossProcessTime() { return cmaxEmbossProcessTime; }
    bool getPhylipnewInstalled() { return cphylipnewInstalled; }
    QString getPrivateKeyFile() { return cprivateKeyFile; }
    QString getCACertsPath() { return cCACertsPath; }
    QString getLocalCertFile() { return clocalCertFile; }

    bool embossProgramExecutionAllowed( const QString &program);

signals:
    void startConnection();

public slots:
    void msgStdOut(const QString &msg);
    void msgStdErr(const QString &msg);
    void threadFinished(QThread* thread);
    void listRunningThreads();
    void killThread(const quint32 thread);
    void shutdownServer() __attribute__((noreturn));

protected:
    void incomingConnection(qintptr socketDescriptor);

private:
    QList<QPointer<KembossThread> > threadList;
    // List of all currently running EMBOSS threads
    quint32 threadID;
    quint16 threadsRunning;

    KInputThread *inputThread;
    // Thread receiving input on stdin and sending signals to the server
    uint cmaxThreadNumber;
    kemboss::serverStatus cserverStatus;
    bool cacceptAllUsers;
    QString cembossFileDir;
    // The directory where the EMBOSS input and output files are saved
    QString cwelcomeMessage;
    // A message that is send to the clients on connection
    QString cembossBinDir;
    // The directory were the EMBOSS programs are installed
    QStringList cembossPrograms;
    // Contains the EMBOSS programs that the server is allowed to execute
    bool cphylipnewInstalled;
    // Is set if the PHYLIPNEW packages are installed on the server
    uint cmaxEmbossProcessTime;
    // The maximum time an EMBOSS program is allowed to run (in secs)
    QString clogfile;
    // Path and filename of the log, the client requests are saved
    QString cprivateKeyFile;
    QString cCACertsPath;
    QString clocalCertFile;
    // The key and certificates files for an encrypted SSL connection
};

#endif // KEMBOSSSERVER_H
