/***************************************************************************
 *                                                                         *
 *   kemboss - Graphical user interface for EMBOSS                         *
 *                                                                         *
 *   Copyright (C) 2007-2014 Tobias Schwarz (mail@tobias-schwarz.net)      *
 *                                                                         *
 *   Based on the EMBOSS.kaptn scripts by Thomas Siegmund                  *
 *                                                                         *
 *   This program is free software; you can redistribute and/or modify     *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version, and provided that the above       *
 *   copyright and permission notice is included with all distributed      *
 *   copies of this or derived software.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 **************************************************************************/

#include "syntax_highlighter.h"

Highlighter::Highlighter(QTextDocument *parent, kembossSettings *fksettings) : QSyntaxHighlighter(parent)
{
   ksettings = fksettings;
   HighlightingRule rule;

   // FASTA sequence name
   rule.pattern = QRegExp("^\\>.*$");
   rule.format = ksettings->seq_name;
   rule.type = noSeq;
   FASTARules.append(rule);

   // FASTA comment
   rule.pattern = QRegExp("^#.*$");
   rule.format = ksettings->comment;
   rule.type = noSeq;
   FASTARules.append(rule);

   // FASTA feature
   rule.pattern = QRegExp("/.*$");
   rule.format = ksettings->comment;
   rule.type = noSeq;
   FASTARules.append(rule);

   // FASTA sequence
   // Everything else is regarded a sequence


   // Alignment title
   rule.pattern = QRegExp("^CLUSTAL.*$");
   rule.format = ksettings->seq_name;
   rule.type = noSeq;
   AlignmentRules.append(rule);

   // Alignment sequence name
   rule.pattern = QRegExp("^\\>.*$");
   rule.format = ksettings->seq_name;
   rule.type = noSeq;
   AlignmentRules.append(rule);

   //Alignment score
   rule.pattern = QRegExp("^(Score|Identities| Plus| Minus).*$");
   rule.format = ksettings->comment;
   rule.type = noSeq;
   AlignmentRules.append(rule);

   // Alignment sequence
   // Everything else is regarded a sequence


   // EMBL sequence name
   rule.pattern = QRegExp("^ID.*$");
   rule.format = ksettings->seq_name;
   rule.type = noSeq;
   EMBLRules.append(rule);

   // EMBL sequence description
   rule.pattern = QRegExp("^DE.*$");
   rule.format = ksettings->seq_name;
   rule.type = noSeq;
   EMBLRules.append(rule);

   // EMBL header
   rule.pattern = QRegExp("^\\w{2}\\s"); // Matches exact 2 word characters at the beginning of the line, followed by a whitespace
   rule.format = ksettings->seq_name;
   rule.type = noSeq;
   EMBLRules.append(rule);

   // EMBL + Gb translation start
   AAStart = QRegExp("/translation=\"");

   // EMBL + Gb translation end
   AAEnd = QRegExp("\"");

   // EMBL any sequence
   rule.pattern = QRegExp("^\\s+[\\w\\s]+\\d+$"); // Matches one ore more whitespaces, one or more word characters or whitespaces, then one or more digits, end of line
   rule.type = anySeq;
   EMBLRules.append(rule);


   // Gb header
   rule.pattern = QRegExp("^\\w+");
   rule.format = ksettings->seq_name;
   rule.type = noSeq;
   GbRules.append(rule);

   // Gb header 2
   rule.pattern = QRegExp("^\\s{2,8}\\w+");
   rule.format = ksettings->comment;
   rule.type = noSeq;
   GbRules.append(rule);

   // Gb any sequence
   rule.pattern = QRegExp("^\\s+\\d+\\s[\\w\\s]+$"); // Matches one ore more whitespaces, then one or more digits, one whitespace, one or more word characters or whitespaces, end of line
   rule.type = anySeq;
   GbRules.append(rule);


   // Swiss header
   rule.pattern = QRegExp("^\\w{2}\\s"); // Matches exact 2 word characters at the beginning of the line, followed by a whitespace
   rule.format = ksettings->seq_name;
   rule.type = noSeq;
   SwissRules.append(rule);

   // Swiss sequence
   rule.pattern = QRegExp("^\\w{3}[\\w\\s]+"); // Matches at least 3 word characters at the beginning of the line, then one or more word characters or whitespaces
   rule.type = anySeq;
   SwissRules.append(rule);


   // AA Sequence rules
   rule.pattern = QRegExp("[cC]+");
   rule.format = ksettings->cystein;
   aaSequenceRules.append(rule);

   rule.pattern = QRegExp("[AVFPMILavfpmil]+");
   rule.format = ksettings->apolar_aa;
   aaSequenceRules.append(rule);

   rule.pattern = QRegExp("[STYHNQWstyhnqtw]+");
   rule.format = ksettings->polar_aa;
   aaSequenceRules.append(rule);

   rule.pattern = QRegExp("[DEde]+");
   rule.format = ksettings->acidic_aa;
   aaSequenceRules.append(rule);

   rule.pattern = QRegExp("[krKR]");
   rule.format = ksettings->basic_aa;
   aaSequenceRules.append(rule);

   rule.pattern = QRegExp("[KRkr]{2,}");
   rule.format = ksettings->dibasic_aa;
   aaSequenceRules.append(rule);


   // Nucleotide sequence rules
   rule.pattern = QRegExp("[aA]+");
   rule.format = ksettings->a_nuc;
   nucSequenceRules.append(rule);

   rule.pattern = QRegExp("[Tt]+");
   rule.format = ksettings->t_nuc;
   nucSequenceRules.append(rule);

   rule.pattern = QRegExp("[gG]+");
   rule.format = ksettings->g_nuc;
   nucSequenceRules.append(rule);

   rule.pattern = QRegExp("[Cc]+");
   rule.format = ksettings->c_nuc;
   nucSequenceRules.append(rule);
}

void Highlighter::setSyntax(const sequenceSyntax newSyntax)
{
   syntax = newSyntax;
}

void Highlighter::setSyntaxInt(const int newSyntax)
{
    switch (newSyntax) {
    case 0:
        syntax = none;
        break;
    case 1:
        syntax = FastaAA;
        break;
    case 2:
        syntax = FastaNuc;
        break;
    case 3:
        syntax = AlignmentAA;
        break;
    case 4:
        syntax = AlignmentNuc;
        break;
    case 5:
        syntax = EMBL;
        break;
    case 6:
        syntax = GenbankAA;
        break;
    case 7:
        syntax = GenbankNuc;
        break;
    case 8:
        syntax = Swiss;
        break;
    }
}

void Highlighter::highlightBlock(const QString &text)
{
    if (syntax == none) {
        QTextCharFormat format;
        setFormat(0, text.length(), format);
        return;
    }

    QVector<HighlightingRule> currentRules;
    if ((syntax == FastaAA) || (syntax == FastaNuc)) {
        currentRules = FASTARules;
    } else if ((syntax == AlignmentAA) || (syntax == AlignmentNuc)) {
        currentRules = AlignmentRules;
    } else if (syntax == EMBL) {
        currentRules = EMBLRules;
    } else if ((syntax == GenbankNuc) || (syntax == GenbankAA)) {
        currentRules = GbRules;
    } else if (syntax == Swiss) {
        currentRules = SwissRules;
    }

    if ((syntax == FastaAA) || (syntax == FastaNuc) || (syntax == AlignmentAA) || (syntax == AlignmentNuc)) {
        bool foundSeq = true;
        foreach (const HighlightingRule &rule, currentRules) {
            QRegExp expression(rule.pattern);
            int index = expression.indexIn(text);
            while (index >= 0) {
                setFormat(index, expression.matchedLength(), rule.format);
                foundSeq = false;
                index = expression.indexIn(text, index + expression.matchedLength());
            }
        }
        if (foundSeq) {
            if ((syntax == FastaAA) || (syntax == AlignmentAA))
                highlightAASequence(text, 0, text.length());
            else if ((syntax == FastaNuc) || (syntax == AlignmentNuc))
                highlightNucSequence(text, 0, text.length());
        }

    } else if ((syntax == GenbankAA) || (syntax == Swiss)) {
        // Genbank or Swissprot AA highlighting
        foreach (const HighlightingRule &rule, currentRules) {
            QRegExp expression(rule.pattern);
            int index = expression.indexIn(text);
            while (index >= 0) {
                if (rule.type == anySeq)
                    highlightAASequence(text, index, index + expression.matchedLength());
                else
                    setFormat(index, expression.matchedLength(), rule.format);
                index = expression.indexIn(text, index + expression.matchedLength());
            }
        }

    } else if ((syntax == EMBL) || (syntax == GenbankNuc)) {
        foreach (const HighlightingRule &rule, currentRules) {
            QRegExp expression(rule.pattern);
            int index = expression.indexIn(text);
            while (index >= 0) {
                if (rule.type == anySeq)
                    highlightNucSequence(text, index, index + expression.matchedLength());
                else
                    setFormat(index, expression.matchedLength(), rule.format);
                index = expression.indexIn(text, index + expression.matchedLength());
            }
        }

        // Store the current block state
        int startIndex;

        if (previousBlockState() == 1) {
            // Previous block ended within an AA sequence
            startIndex = 2;
        } else {
            // Previous block did not end within an AA sequence
            startIndex = AAStart.indexIn(text); // Returns -1 if no AA start is found within the text block
            if (startIndex >= 0)
                startIndex += 14; // The AA sequence starts after the expression /translation="
            else
                setCurrentBlockState(0);
        }

        while (startIndex >= 2) {
            // The text block contains an AA sequence. We have to lock for an AA sequence end
            int endIndex = AAEnd.indexIn(text, startIndex);
            if (endIndex == -1) {
                // No end of the AA sequence found. The current block ends with an AA sequence
                setCurrentBlockState(1);
                endIndex = text.length();
            } else {
                setCurrentBlockState(0);
            }
            highlightAASequence(text, startIndex, endIndex);
            startIndex = AAStart.indexIn(text, startIndex + 1);
        }
    }
}

void Highlighter::highlightAASequence(const QString &text, const int start, const int end)
{
  foreach (const HighlightingRule &rule, aaSequenceRules) {
      QRegExp expression(rule.pattern);
      int index = expression.indexIn(text);
      while ((index >= 0) && (index < end)) {
         if (index >=start) setFormat(index, expression.matchedLength(), rule.format);
         index = expression.indexIn(text, index + expression.matchedLength());
      }
   }
}

void Highlighter::highlightNucSequence(const QString &text, const int start, const int end)
{
   foreach (const HighlightingRule &rule, nucSequenceRules) {
      QRegExp expression(rule.pattern);
      int index = expression.indexIn(text);
      while ((index >= 0) && (index < end)) {
         if (index >=start) setFormat(index, expression.matchedLength(), rule.format);
         index = expression.indexIn(text, index + expression.matchedLength());
      }
   }
}
